@@ -1,32 +1,76 @@
 package Pegex::Parser;
 use Pegex::Base;
+no warnings qw( recursion );
 
 use Pegex::Input;
 use Pegex::Optimizer;
 use Scalar::Util;
 
-# XXX Need to figure out what uses this. (sample.t)
-{
-    package Pegex::Constant;
-    our $Null = [];
-    our $Dummy = [];
-}
-
 has grammar => (required => 1);
-has 'debug' => ();
+has receiver => ();
+has input => ();
+
+has recursion_count => 0;
+has iteration_count => 0;
+
+has debug => ();
+has debug_indent => ();
+has debug_color => ();
+has debug_got_color => ();
+has debug_not_color => ();
+
+has recursion_limit => ();
+has recursion_warn_limit => ();
+has iteration_limit => ();
+
 sub BUILD {
     my ($self) = @_;
-    $self->{receiver} ||= undef;
-    $self->{input} ||= undef;
-    $self->{rule} = undef;
-    $self->{parent} = undef;
-    $self->{position} = 0;
-    $self->{farthest} = 0;
-    $self->{debug} =
-        exists($ENV{PERL_PEGEX_DEBUG}) ? $ENV{PERL_PEGEX_DEBUG} :
-        defined($Pegex::Parser::Debug) ? $Pegex::Parser::Debug :
-        0;
+
     $self->{throw_on_error} ||= 1;
+
+    $self->{debug} //=
+        $ENV{PERL_PEGEX_DEBUG} //
+        $Pegex::Parser::Debug // 0;
+
+    $self->{debug_indent} //=
+        $ENV{PERL_PEGEX_DEBUG_INDENT} //
+        $Pegex::Parser::DebugIndent // 1;
+    $self->{debug_indent} = 1 if (
+        not length $self->{debug_indent}
+        or $self->{debug_indent} =~ tr/0-9//c
+        or $self->{debug_indent} < 0
+    );
+
+    {
+        $self->{debug_color} //=
+            $ENV{PERL_PEGEX_DEBUG_COLOR} //
+            $Pegex::Parser::DebugColor // $self->{debug};
+        my ($got, $not);
+        ($self->{debug_color}, $got, $not) = split / *, */, $self->{debug_color};
+        $got ||= 'bright_green';
+        $not ||= 'bright_red';
+        $_ = [split ' ', $_] for ($got, $not);
+        $self->{debug_got_color} = $got;
+        $self->{debug_not_color} = $not;
+        my $c = $self->{debug_color} // $self->{debug};
+        $self->{debug_color} =
+            $c eq 'on' ? 1 :
+            $c eq 'auto' ? (-t STDERR ? 1 : 0) :
+            $c eq 'off' ? 0 :
+            $c =~ /^\d+$/ ? $c : 0;
+        $self->{debug_color} and
+            eval {require v5.14; require Term::ANSIColor; 1} or
+            $self->{debug_color} = 0;
+    }
+    $self->{recursion_limit} //=
+        $ENV{PERL_PEGEX_RECURSION_LIMIT} //
+        $Pegex::Parser::RecursionLimit // 0;
+    $self->{recursion_warn_limit} //=
+        $ENV{PERL_PEGEX_RECURSION_WARN_LIMIT} //
+        $Pegex::Parser::RecursionWarnLimit // 0;
+    $self->{iteration_limit} //=
+        $ENV{PERL_PEGEX_ITERATION_LIMIT} //
+        $Pegex::Parser::IterationLimit // 0;
 }
 
 # XXX Add an optional $position argument. Default to 0. This is the position
@@ -48,9 +92,8 @@ sub parse {
     $self->{input}->open
         unless $self->{input}{_is_open};
     $self->{buffer} = $self->{input}->read;
-
-    die "No 'grammar'. Can't parse"
-        unless $self->{grammar};
+    $self->{last_line_pos} = 0;
+    $self->{last_line} = 1;
 
     $self->{grammar}{tree} ||= $self->{grammar}->make_tree;
 
@@ -80,6 +123,17 @@ sub parse {
         $self->{receiver}->initial();
     }
 
+    local *match_next;
+    {
+        no warnings 'redefine';
+        *match_next = (
+            $self->{recursion_warn_limit} or
+            $self->{recursion_limit} or
+            $self->{iteration_limit}
+         ) ? \&match_next_with_limit :
+             \&match_next_normal;
+    }
+
     my $match = $self->debug ? do {
         my $method = $optimizer->make_trace_wrapper(\&match_ref);
         $self->$method($start_rule_ref, {'+asr' => 0});
@@ -101,7 +155,7 @@ sub parse {
     $match->[0];
 }
 
-sub match_next {
+sub match_next_normal {
     my ($self, $next) = @_;
 
     my ($rule, $method, $kind, $min, $max, $assertion) =
@@ -126,8 +180,6 @@ sub match_next {
         else {
             $match = [$match]
         }
-        $self->{farthest} = $position
-            if ($self->{position} = $position) > $self->{farthest};
     }
     my $result = ($count >= $min and (not $max or $count <= $max))
         ^ ($assertion == -1);
@@ -139,6 +191,36 @@ sub match_next {
     ($result ? $next->{'-skip'} ? [] : $match : 0);
 }
 
+sub match_next_with_limit {
+    my ($self, $next) = @_;
+
+    sub limit_msg {
+        "Deep recursion ($_[0] levels) on Pegex::Parser::match_next\n";
+    }
+
+    $self->{iteration_count}++;
+    $self->{recursion_count}++;
+
+    if (
+        $self->{recursion_limit} and
+        $self->{recursion_count} >= $self->{recursion_limit}
+    ) { die limit_msg $self->{recursion_count} }
+    elsif (
+        $self->{recursion_warn_limit} and
+        not ($self->{recursion_count} % $self->{recursion_warn_limit})
+    ) { warn limit_msg $self->{recursion_count} }
+    elsif (
+        $self->{iteration_limit} and
+        $self->{iteration_count} > $self->{iteration_limit}
+    ) { die "Pegex iteration limit of $self->{iteration_limit} reached." }
+
+    my $result = $self->match_next_normal($next);
+
+    $self->{recursion_count}--;
+
+    return $result;
+}
+
 sub match_rule {
     my ($self, $position, $match) = (@_, []);
     $self->{position} = $position;
@@ -226,7 +308,16 @@ sub trace {
     my $indent = ($action =~ /^try_/) ? 1 : 0;
     $self->{indent} ||= 0;
     $self->{indent}-- unless $indent;
-    print STDERR ' ' x $self->{indent};
+
+    $action = (
+        $action =~ m/got_/ ?
+            Term::ANSIColor::colored($self->{debug_got_color}, $action) :
+        $action =~ m/not_/ ?
+            Term::ANSIColor::colored($self->{debug_not_color}, $action) :
+        $action
+    ) if $self->{debug_color};
+
+    print STDERR ' ' x ($self->{indent} * $self->{debug_indent});
     $self->{indent}++ if $indent;
     my $snippet = substr(${$self->{buffer}}, $self->{position});
     $snippet = substr($snippet, 0, 30) . "..."
@@ -250,7 +341,7 @@ sub format_error {
     my $position = $self->{farthest};
     my $real_pos = $self->{position};
 
-    my $line = @{[substr($$buffer, 0, $position) =~ /(\n)/g]} + 1;
+    my $line = $self->line($position);
     my $column = $position - rindex($$buffer, "\n", $position);
 
     my $pretext = substr(
@@ -278,9 +369,37 @@ sub line_column {
     my ($self, $position) = @_;
     $position ||= $self->{position};
     my $buffer = $self->{buffer};
-    my $line = @{[substr($$buffer, 0, $position) =~ /(\n)/g]} + 1;
+    my $line = $self->line($position);
     my $column = $position - rindex($$buffer, "\n", $position);
-    return [$line, $position];
+    return [$line, $column];
+}
+
+sub line {
+    my ($self, $position) = @_;
+    $position ||= $self->{position};
+    my $buffer = $self->{buffer};
+    my $last_line = $self->{last_line};
+    my $last_line_pos = $self->{last_line_pos};
+    my $len = $position - $last_line_pos;
+    if ($len == 0) {
+        return $last_line;
+    }
+    my $line;
+    if ($len < 0) {
+        $line = $last_line - scalar substr($$buffer, $position, -$len) =~ tr/\n//;
+    } else {
+        $line = $last_line + scalar substr($$buffer, $last_line_pos, $len) =~ tr/\n//;
+    }
+    $self->{last_line} = $line;
+    $self->{last_line_pos} = $position;
+    return $line;
+}
+
+# XXX Need to figure out what uses this. (sample.t)
+{
+    package Pegex::Constant;
+    our $Null = [];
+    our $Dummy = [];
 }
 
 1;
